# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import Lldp
from netq_lib.common.enums import ServiceStatus


def parse_cl_lldp(lldp):
    lldpobj = Lldp()
    lldpobj.ifname = str(lldp['name'])
    try:
        lldpobj.peer_hostname = str(
            lldp['chassis'][0]['name'][0]['value']
        )
    except:
        # chassis id will always be present
        lldpobj.peer_hostname = (
            str(lldp['chassis'][0]['id'][0]['type']) + ":" +
            str(lldp['chassis'][0]['id'][0]['value'])
        )

    if '.' in lldpobj.peer_hostname:
        lldpobj.peer_hostname = (
            lldpobj.peer_hostname.split('.')[0]
        )
    if lldp['port'][0]['id'][0]['type'] == 'ifname':
        lldpobj.peer_ifname = str(lldp['port'][0]['id'][0]['value'])
    else:
        # port id will be present always - sometimes as mac sub-type
        lldpobj.peer_ifname = (str(lldp['port'][0]['id'][0]['type']) + ":" +
                               str(lldp['port'][0]['id'][0]['value']))

    try:
        descr = lldp['chassis'][0]['descr'][0]['value']
    except:
        descr = 'unknown'

    words = descr.split()
    if words[0] == 'Cumulus':
        if len(words) > 3:
            lldpobj.lldp_peer_os = (
                ' '.join(words[:2]).encode('ascii', 'ignore'))
            if words[2] == 'version':
                try:
                    if "(" in words[3]:
                        lldpobj.lldp_peer_osv = (
                            words[4].split(')')[1].encode('ascii',
                                                  'ignore'))
                    else:
                        lldpobj.lldp_peer_osv = (
                            words[3].encode('ascii', 'ignore'))
                except:
                    lldpobj.lldp_peer_osv = (
                        ' '.join(words[3:4]).encode('ascii', 'ignore'))
            else:
                lldpobj.lldp_peer_osv = 'unknown'
        else:
            lldpobj.lldp_peer_os = words[0].encode('ascii', 'ignore')
            lldpobj.lldp_peer_osv = 'unknown'
    elif words[0] == 'Ubuntu':
        lldpobj.lldp_peer_os = words[0].encode('ascii',
                                               'ignore')
        if len(words) > 1:
            if "16" in words[1]:
                lldpobj.lldp_peer_osv = words[1].encode('ascii',
                                                        'ignore')
            else:
                lldpobj.lldp_peer_osv = 'unknown'
        else:
            lldpobj.lldp_peer_osv = 'unknown'
    else:
        lldpobj.lldp_peer_os = words[0].encode('ascii', 'ignore')
        lldpobj.lldp_peer_osv = 'unknown'

    try:
        for cap in lldp['chassis'][0]['capability']:
            if cap['type'] == 'Bridge':
                lldpobj.lldp_peer_bridge = cap['enabled']
            elif cap['type'] == 'Router':
                lldpobj.lldp_peer_router = cap['enabled']
            elif cap['type'] == 'Station':
                lldpobj.lldp_peer_station = cap['enabled']
    except:
        pass

    lldpobj.save()
    return lldpobj


def parse_capability(capability_list):
    capability = {}

    if not isinstance(capability_list, list):
        return capability

    for cap in capability_list:
        if cap['type'] == 'Bridge':
            capability['bridge'] = cap['enabled']
        elif cap['type'] == 'Router':
            capability['router'] = cap['enabled']
        elif cap['type'] == 'Station':
            capability['station'] = cap['enabled']

    return capability


def parse_os_info(descr):
    words = descr.split()
    peer_os = ''
    peer_osv = ''

    if words[0] == 'Cumulus':
        peer_os = (' '.join(words[:2]).encode('ascii', 'ignore'))
        try:
            if "(" in words[3]:
                peer_osv = (words[4].split(')')[1].encode('ascii',
                                                          'ignore'))
            else:
                peer_osv = (words[3].encode('ascii', 'ignore'))
        except:
            peer_osv = (' '.join(words[3:4]).encode('ascii', 'ignore'))
    elif words[0] == 'Ubuntu':
        peer_os = words[0].encode('ascii', 'ignore')
        peer_osv = words[1].encode('ascii', 'ignore')
    return (peer_os, peer_osv)


def parse_host_lldp(lldp):
    lldpobj = Lldp()
    #
    # Walk through each interface
    #
    for ifname, lldp_info in lldp.iteritems():
        lldpobj.ifname = str(ifname)

        for key, value in lldp_info.iteritems():
            if key == 'chassis':
                for peer, peer_info in value.iteritems():
                    lldpobj.peer_hostname = str(peer)
                    for peer_attr, peer_attr_value in peer_info.iteritems():
                        if peer_attr == 'descr':
                            peer_os, peer_osv = parse_os_info(peer_attr_value)
                            lldpobj.lldp_peer_osv = peer_osv
                            lldpobj.lldp_peer_os = peer_os
                        if peer_attr == 'capability':
                            capability = parse_capability(peer_attr_value)
                            lldpobj.lldp_peer_bridge = capability.get('bridge',
                                                                      False)
                            lldpobj.lldp_peer_router = capability.get('router',
                                                                      False)
                            lldpobj.lldp_peer_station = capability.get(
                                'station', False)
            if key == 'port':
                port_id = value.get('id', {})
                lldpobj.peer_ifname = str(port_id.get('value'))

    lldpobj.save()
    return lldpobj


def parse_lldp(lldp, platform):
    if platform and 'ubuntu' in platform.lower():
        new_lldp_ent = parse_host_lldp(lldp)
    elif platform and 'centos' in platform.lower():
        new_lldp_ent = parse_host_lldp(lldp)
    else:
        new_lldp_ent = parse_cl_lldp(lldp)
    return new_lldp_ent


def process_lldpctl(cmd_output, svc_failed=False, **kwargs):
    '''
    Process LLDPD's lldpctl output and push appropriate LLDP data to DB
    '''
    retcode = ServiceStatus.OK
    platform = kwargs.get('platform')

    try:
        cmd_output = cmd_output.decode('string-escape')
        json_out = json.loads(cmd_output)
    except ValueError:
        m_lldp = Lldp.memmodel()
        for entry in m_lldp.query.filter(hostname=HOSTNAME):
            old_lldp_ent = Lldp()
            old_lldp_ent.copy(entry)
            old_lldp_ent.delete()
        return ServiceStatus.FAIL

    # grab current list of lldp objects
    m_lldp = Lldp.memmodel()
    old_lldp_objs = []
    new_lldp_objs = []
    if m_lldp:
        old_lldp_objs = (
            list(m_lldp.query.filter(hostname=HOSTNAME))
        )

    for values in json_out.itervalues():
        if isinstance(values, list):
            for value in values:
                if not value.get('interface', None):
                    continue
                if isinstance(value['interface'], list):
                    for lldp in value['interface']:
                        new_lldp_ent = parse_lldp(lldp, platform)
                        new_lldp_objs.append(new_lldp_ent)
                else:
                    new_lldp_ent = parse_lldp(value['interface'], platform)
                    new_lldp_objs.append(new_lldp_ent)
        else:
            if not values.get('interface', None):
                continue
            if isinstance(values['interface'], list):
                for lldp in values['interface']:
                    new_lldp_ent = parse_lldp(lldp, platform)
                    new_lldp_objs.append(new_lldp_ent)
            else:
                new_lldp_ent = parse_lldp(values['interface'], platform)
                new_lldp_objs.append(new_lldp_ent)

    # remove current elements from old list
    for new_ele in new_lldp_objs:
        for old_ele in old_lldp_objs:
           if old_ele.equals(new_ele):
               old_lldp_objs.remove(old_ele)

    # delete remaining stale entries
    for old_ele in old_lldp_objs:
       old_lldp_ent = Lldp()
       old_lldp_ent.copy(old_ele)
       old_lldp_ent.delete()

    return retcode
