# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

""" This module provides classes with methods to handle netlink notifications.
"""
from __future__ import absolute_import

import collections
import os
import socket
import struct
import sys

import greenlet
import eventlet
import eventlet.pools
from nlmanager import nlpacket
from nlmanager.nlpacket import (
    NLM_F_DUMP,
    NLM_F_REQUEST,
    RTM_DELADDR,
    RTM_DELLINK,
    RTM_DELROUTE,
    RTM_DELNEIGH,
    RTM_GETADDR,
    RTM_GETNEIGH,
    RTM_GETROUTE,
    RTM_GETLINK,
    RTM_NEWADDR,
    RTM_NEWLINK,
    RTM_NEWNEIGH,
    RTM_NEWROUTE,
    RTMGRP_IPV4_IFADDR,
    RTMGRP_IPV6_ROUTE,
    RTMGRP_IPV6_IFADDR,
    RTMGRP_IPV4_ROUTE,
    RTMGRP_LINK,
    RTMGRP_NEIGH,
    padded_length
)

from netq_lib.common.enums import (
    InterfaceState,
    InterfaceDownReason,
    LinkType,
    NeighborState,
    RouteType,
    RouteTable,
    Protocol,
    StpPortState
)

from netq_lib.common.utils import is_valid_mac


class Address(object):

    def __init__(self, ifindex, prefix, mask, family):
        self.ifindex = ifindex
        self.prefix = prefix
        self.mask = mask
        self.is_ipv6 = family == socket.AF_INET6


class Link(object):

    def __init__(self, **kwargs):
        self.family = kwargs['family']
        self.ifindex = kwargs['ifindex']
        self.ifname = kwargs['ifname']
        self.admin_state = kwargs['admin_state']
        self.oper_state = kwargs['oper_state']
        self.mac_addr = kwargs['mac']
        kind = kwargs.get('kind', None)
        if kind == 'ignore':
            self.kind = ''
        else:
            self.kind = kind
        self.mtu = kwargs.get('mtu', 1500)
        self.rt_table_id = kwargs.get('rt_table_id', 0)
        self.master_ifindex = kwargs['master']
        self.dstport = kwargs.get('dstport', 0)
        self.localip = kwargs.get('localip', '')
        self.vni = int(kwargs.get('vni', 0))
        self.is_vlan_filtering = kwargs.get('is_vlan_filtering', False)
        self.parent_if = kwargs.get('parent_if', None)
        self.vlans = kwargs.get('vlans', '')
        self.access_vlan = kwargs.get('access_vlan', 0)
        self.stp_state = kwargs.get('stp_state', StpPortState.FORWARDING)
        self.down_reason = kwargs.get('down_reason', InterfaceDownReason.NONE)
        self.learning_en = kwargs.get('learning_en', False)
        self.arp_suppress_en = kwargs.get('arp_suppress_en', False)
        self.ifalias = kwargs.get('ifalias', '')


class Neighbor(object):

    def __init__(self, **kwargs):
        self.ifindex = kwargs.get('ifindex')
        self.is_ipv6 = kwargs.get('family') == socket.AF_INET6
        self.ip_address = str(kwargs.get('ip'))
        self.mac_address = kwargs.get('mac')
        self.state = kwargs.get('state')
        self.family = kwargs.get('family')
        self.vlan = kwargs.get('vlan')
        self.port = kwargs.get('port')
        self.origin = kwargs.get('origin')
        self.flags = kwargs.get('flags')
        self.is_static = kwargs.get('is_static')


class Route(object):

    def __init__(self, origin, ip_address, nexthops, protocol, route_type,
                 src, family, table):
        self.ifindex = Netlink.DUMMY_IFINDEX
        self.origin = origin
        self.ip_address = ip_address
        self.is_ipv6 = family == socket.AF_INET6
        self.nexthops = nexthops
        self.protocol = protocol
        self.route_type = route_type
        self.src = src
        self.table = table

    @staticmethod
    def get_nexthops(msg):
        nexthop = msg.get_attribute_value(msg.RTA_GATEWAY)
        multipath = msg.get_attribute_value(msg.RTA_MULTIPATH)
        nexthops = []
        if nexthop:
            rta_oif = msg.get_attribute_value(msg.RTA_OIF)
            nexthops.append((nexthop, rta_oif))
        elif multipath:
            for (nexthop, rtnh_ifindex, _, _) in multipath:
                nexthops.append((nexthop, rtnh_ifindex))
        else:
            rta_oif = msg.get_attribute_value(msg.RTA_OIF)
            nexthops.append((None, rta_oif))
        return nexthops


class NetlinkStat(object):

    RTM_NEWFDB = 'RTM_NEWFDB'
    RTM_DELFDB = 'RTM_DELFDB'

    str_map = {
        RTM_NEWLINK: 'RTM_NEWLINK',
        RTM_DELLINK: 'RTM_DELLINK',
        RTM_NEWNEIGH: 'RTM_NEWNEIGH',
        RTM_DELNEIGH: 'RTM_DELNEIGH',
        RTM_NEWROUTE: 'RTM_NEWROUTE',
        RTM_DELROUTE: 'RTM_DELROUTE',
        RTM_NEWADDR: 'RTM_NEWADDR',
        RTM_DELADDR: 'RTM_DELADDR',
        RTM_NEWFDB: 'RTM_NEWFDB',
        RTM_DELFDB: 'RTM_DELFDB'
    }

    ALL = [RTM_NEWLINK, RTM_DELLINK, RTM_NEWNEIGH, RTM_DELNEIGH,
           RTM_NEWROUTE, RTM_DELROUTE, RTM_NEWADDR, RTM_DELADDR,
           RTM_NEWFDB, RTM_DELFDB]

    def __init__(self):
        self.stats = None

    def __repr__(self):
        return repr({self.str_map[k]: v for k, v in self.stats.iteritems()})

    def increment(self, msg_type, count):
        self.stats[msg_type] += count

    def reset(self):
        self.stats = {stat: 0 for stat in NetlinkStat.ALL}


class Netlink(object):
    """ Provides methods to parse RTNL messages and invoke callbacks based on
    the message type.
    """
    # pylint: disable=too-many-instance-attributes
    __BUF_SIZE = 90 * 1024 * 1024
    __NETLINK_NO_ENOBUFS = 5
    __NLMSG_FMT = 'IHHII'
    __NLMSG_LEN = struct.calcsize(__NLMSG_FMT)
    __SO_RCVBUFFORCE = 33
    __SOL_NETLINK = 270

    DUMMY_IFINDEX = 0
    GROUPS = (RTMGRP_LINK | RTMGRP_NEIGH | RTMGRP_IPV4_IFADDR |
              RTMGRP_IPV4_ROUTE | RTMGRP_IPV6_IFADDR | RTMGRP_IPV6_ROUTE)
    NETLINK_ROUTE = 0            # Routing/device hook
    NLSOCK_BYTES = 8 * 1024

    NEWADDR = RTM_NEWADDR
    DELADDR = RTM_DELADDR
    NEWLINK = RTM_NEWLINK
    DELLINK = RTM_DELLINK
    NEWNEIGH = RTM_NEWNEIGH
    DELNEIGH = RTM_DELNEIGH
    NEWROUTE = RTM_NEWROUTE
    DELROUTE = RTM_DELROUTE

    IFA_ADDRESS = nlpacket.Address.IFA_ADDRESS
    IFF_SLAVE = nlpacket.Link.IFF_SLAVE
    IFLA_ADDRESS = nlpacket.Link.IFLA_ADDRESS
    IFLA_IFNAME = nlpacket.Link.IFLA_IFNAME
    IFLA_IFALIAS = nlpacket.Link.IFLA_IFALIAS
    IFLA_LINK = nlpacket.Link.IFLA_LINK
    IFLA_LINKINFO = nlpacket.Link.IFLA_LINKINFO
    IFLA_MASTER = nlpacket.Link.IFLA_MASTER
    IFLA_CARRIER = nlpacket.Link.IFLA_CARRIER
    IFLA_MTU = nlpacket.Link.IFLA_MTU
    IFLA_INFO_KIND = nlpacket.Link.IFLA_INFO_KIND
    IFLA_INFO_DATA = nlpacket.Link.IFLA_INFO_DATA
    IFLA_INFO_SLAVE_KIND = nlpacket.Link.IFLA_INFO_SLAVE_KIND
    IFLA_INFO_SLAVE_DATA = nlpacket.Link.IFLA_INFO_SLAVE_DATA
    IFLA_VLAN_ID = nlpacket.Link.IFLA_VLAN_ID
    IFLA_VXLAN_GROUP = nlpacket.Link.IFLA_VXLAN_GROUP
    IFLA_VXLAN_ID = nlpacket.Link.IFLA_VXLAN_ID
    IFLA_VXLAN_LOCAL = nlpacket.Link.IFLA_VXLAN_LOCAL
    IFLA_VXLAN_PORT = nlpacket.Link.IFLA_VXLAN_PORT
    IFLA_VXLAN_LEARNING = nlpacket.Link.IFLA_VXLAN_LEARNING
    IFLA_AF_SPEC = nlpacket.Link.IFLA_AF_SPEC
    IFLA_PROTINFO = nlpacket.Link.IFLA_PROTINFO
    IFLA_BRIDGE_VLAN_INFO = nlpacket.Link.IFLA_BRIDGE_VLAN_INFO
    IFLA_BRPORT_STATE = nlpacket.Link.IFLA_BRPORT_STATE
    IFLA_BRPORT_LEARNING = nlpacket.Link.IFLA_BRPORT_LEARNING
    IFLA_BRPORT_ARP_SUPPRESS = nlpacket.Link.IFLA_BRPORT_ARP_SUPPRESS
    BRIDGE_VLAN_INFO_PVID = nlpacket.Link.BRIDGE_VLAN_INFO_PVID
    BRIDGE_VLAN_INFO_RANGE_BEGIN = nlpacket.Link.BRIDGE_VLAN_INFO_RANGE_BEGIN
    BRIDGE_VLAN_INFO_RANGE_END = nlpacket.Link.BRIDGE_VLAN_INFO_RANGE_END
    try:
        IFLA_BR_VLAN_FILTERING = nlpacket.Link.IFLA_BR_VLAN_FILTERING
        IFLA_BR_TOPOLOGY_CHANGE = nlpacket.Link.IFLA_BR_TOPOLOGY_CHANGE
        IFLA_BR_TOPOLOGY_CHANGE_DETECTED = \
            nlpacket.Link.IFLA_BR_TOPOLOGY_CHANGE_DETECTED
    except:
        IFLA_BR_VLAN_FILTERING = 7
        IFLA_BR_TOPOLOGY_CHANGE = 14
        IFLA_BR_TOPOLOGY_CHANGE_DETECTED = 15

    IFLA_VRF_TABLE = 1                    # Not defined elsewhere
    IFLA_VRF_PORT_TABLE = 1               # Not defined elsewhere
    NDA_DST = nlpacket.Neighbor.NDA_DST
    NDA_LLADDR = nlpacket.Neighbor.NDA_LLADDR
    NDA_PROBES = nlpacket.Neighbor.NDA_PROBES
    # This should eventually make its way into nlmananger
    NDA_VLAN = NDA_PROBES + 1
    NDA_PORT = NDA_PROBES + 2
    NDA_VNI = NDA_PROBES + 3
    NDA_IFINDEX = NDA_PROBES + 4
    NDA_MASTER = NDA_PROBES + 5
    RTA_PREFSRC = nlpacket.Route.RTA_PREFSRC
    RTA_DST = nlpacket.Route.RTA_DST
    RTA_TABLE = nlpacket.Route.RTA_TABLE
    VXLAN_ID = nlpacket.Link.IFLA_VXLAN_ID
    RTA_TABLE = nlpacket.Route.RTA_TABLE
    NUD_PERMANENT = nlpacket.Neighbor.NUD_PERMANENT
    NUD_STALE = nlpacket.Neighbor.NUD_STALE
    NUD_REACHABLE = nlpacket.Neighbor.NUD_REACHABLE
    NUD_FAILED = nlpacket.Neighbor.NUD_FAILED
    NUD_INCOMPLETE = nlpacket.Neighbor.NUD_INCOMPLETE
    NUD_DELAY = nlpacket.Neighbor.NUD_DELAY
    NUD_NOARP = nlpacket.Neighbor.NUD_NOARP

    def __init__(self, proc_name, conf, logger, process_cb, pool=None):
        # debug is a reloadable option; so we need to store a reference to
        # the configuration object and inspect the debug value at runtime
        self.__conf = conf
        self.__intfevent = None
        self.__intfqueue = None
        self.__logger = logger
        self.__debug = False
        # Turning on "debug" here shows the agent as Rotten.
        # if self.__logger.isEnabledFor(logging.DEBUG):
        #    self.__debug = True
        self.__unacknowledged = None
        # PID_MAX_LIMIT is 2^22 allowing 1024 sockets per-pid. We
        # start with 1 in the upper space (top 10 bits) instead of
        # 0 to avoid conflicts with netlink_autobind which always
        # attempts to bind with the pid (and on failure with
        # negative values -4097, -4098, -4099 etc.)
        self.__pid = os.getpid() | (1 << 22)
        self.__pool = pool or eventlet.GreenPool()
        self.__process_cbs = process_cb
        self.__running = set()
        self.__tsocketpool = eventlet.pools.TokenPool(max_size=1)
        self.__seqno = None
        self.dispatcher_gt = None
        self.stats = NetlinkStat()
        self.socket = None

        # parse config file for options
        self.__disable_route_update = False
        self.__disable_mac_nbr_update = False
        if self.__conf.get(proc_name, None):
             # reinit netlink groups for bind call
             self.GROUPS = (RTMGRP_LINK | RTMGRP_IPV4_IFADDR | RTMGRP_IPV6_IFADDR)
             self.__disable_route_update = conf[proc_name].get('disable_route_update', False)

             if not self.__disable_route_update:
                self.GROUPS |= (RTMGRP_IPV4_ROUTE | RTMGRP_IPV6_ROUTE)
             else:
                self.__logger.info('Disable netlink ROUTE updates')

             self.__disable_mac_nbr_update = conf[proc_name].get('disable_mac_nbr_update', False)
             if not self.__disable_mac_nbr_update:
                self.GROUPS |= (RTMGRP_NEIGH)
             else:
                self.__logger.info('Disable netlink NBR/MAC updates')

        self.reset()

    def __dispatcher(self):
        """ Main thread that drains interface queues by invoking handler
        methods based on the message type.
        Netlink messages for different interfaces can be served concurrently;
        however, strict ordering must be maintained between messages
        for the same interface. In order to facilitate this, a queue is
        allocated on a per interface basis, and the dispatcher uses a
        modified FCFS algorithm to schedule threads for messages in these
        queues.
        Dispatch algorithm:
        The dispatcher monitors the intfevent queue for interface events. On
        receiving a new event, it checks to see if another thread is
        serving the same interface. If not, it spawns one to handle the first
        message in the interface's queue. When the new thread is done, it
        notifies the dispatcher by placing the interface's index on the
        "intfevent" queue.
        """
        while True:
            ifindex = self.__intfevent.get()
            if ifindex in self.__intfqueue and ifindex not in self.__running:
                try:
                    msg_type, info = self.__intfqueue[ifindex].get_nowait()
                except eventlet.queue.Empty:
                    self.__intfqueue.pop(ifindex)
                else:
                    self.__running.add(ifindex)
                    green_thread = self.__pool.spawn(
                        self.__process_cbs[msg_type],
                        info
                    )
                    green_thread.link(self.__stop_checker)

    @staticmethod
    def __get_value(attrs, key):
        """ Returns the value associated with a key in attrs.
        """
        return next((val for attr, val in attrs or [] if attr == key), None)

    def __stop_checker(self, green_thread):
        """ Propagates exceptions to the dispatcher green thread.
        """
        ifindex = None
        try:
            ifindex = green_thread.wait()
        except greenlet.GreenletExit:  # pylint: disable=no-member
            pass
        except Exception:  # pylint: disable=broad-except
            eventlet.kill(self.dispatcher_gt, *sys.exc_info())
        self.__running.remove(ifindex)
        self.__intfevent.put(ifindex)

    def __tx_nlpacket(self, pkt):
        """TX the message and wait for an ack
        """
        # NetlinkListener looks at the manager's target_seq and target_pid
        # to know when we've RXed the ack that we want
        identifier = (pkt.seq, pkt.pid)
        self.__unacknowledged.add(identifier)
        interval = 1
        while identifier in self.__unacknowledged:
            with self.__tsocketpool.item() as _:
                try:
                    self.socket.sendall(pkt.message)
                except Exception:  # pylint: disable=broad-except
                    self.__logger.exception('Failed to send netlink message')
            eventlet.sleep(interval)
            interval *= 2
        return True

    def bind(self):
        """ Binds the netlink socket and initializes data structures used for
        communication between the dispatcher and the server.
        """
        with self.__tsocketpool.item() as _:
            if self.socket is not None:
                try:
                    self.socket.close()
                except socket.error as ex:
                    # Ignore the error as we will try to rebind.
                    self.__logger.warning('socket close failed %s', ex)
            try:
                # pylint: disable=no-member
                self.socket = socket.socket(socket.AF_NETLINK,
                                            socket.SOCK_RAW,
                                            self.NETLINK_ROUTE)
                # Set rcv. buffer size to 30M (higher than the rmem_max of 8M).
                self.socket.setsockopt(socket.SOL_SOCKET,
                                       self.__SO_RCVBUFFORCE,
                                       self.__BUF_SIZE)
            except socket.error as ex:
                self.__logger.warning('socket open failed %s', ex)
                raise RuntimeError('open: socket err: %s' % ex)
            # Open a socket for receiving netlink msgs.
            try:
                self.socket.bind((self.__pid, self.GROUPS))
            except socket.error as ex:
                self.__logger.warning('socket bind failed %s', ex)
                raise RuntimeError('bind: socket err: %s' % ex)

    def get_all_addresses(self):
        family = socket.AF_UNSPEC
        pkt = nlpacket.Address(RTM_GETADDR, self.__debug, self.__logger)
        pkt.flags = NLM_F_REQUEST | NLM_F_DUMP
        pkt.body = struct.pack('Bxxxi', family, 0)
        pkt.build_message(self.__seqno.next(), self.__pid)
        return self.__tx_nlpacket(pkt)

    def get_all_links(self):
        family = socket.AF_UNSPEC
        pkt = nlpacket.Link(RTM_GETLINK, self.__debug, self.__logger)
        pkt.flags = NLM_F_REQUEST | NLM_F_DUMP
        pkt.body = struct.pack('Bxxxiii', family, 0, 0, 0)
        pkt.build_message(self.__seqno.next(), self.__pid)
        self.__tx_nlpacket(pkt)
        family = socket.AF_BRIDGE
        pkt.body = struct.pack('Bxxxiii', family, 0, 0, 0)
        pkt.add_attribute(nlpacket.Link.IFLA_EXT_MASK,
                          nlpacket.Link.RTEXT_FILTER_BRVLAN_COMPRESSED)
        pkt.build_message(self.__seqno.next(), self.__pid)
        return self.__tx_nlpacket(pkt)

    def get_all_neighbors(self):
        if self.__disable_mac_nbr_update:
            return True
        family = socket.AF_UNSPEC
        pkt = nlpacket.Neighbor(RTM_GETNEIGH, self.__debug, self.__logger)
        pkt.flags = NLM_F_REQUEST | NLM_F_DUMP
        pkt.body = struct.pack('Bxxxii', family, 0, 0)
        pkt.build_message(self.__seqno.next(), self.__pid)
        return self.__tx_nlpacket(pkt)

    def get_all_macs(self):
        if self.__disable_mac_nbr_update:
            return True
        family = socket.AF_BRIDGE
        pkt = nlpacket.Neighbor(RTM_GETNEIGH, self.__debug, self.__logger)
        pkt.flags = NLM_F_REQUEST | NLM_F_DUMP
        pkt.body = struct.pack('Bxxxii', family, 0, 0)
        pkt.build_message(self.__seqno.next(), self.__pid)
        return self.__tx_nlpacket(pkt)

    def get_all_routes(self):
        if self.__disable_route_update:
            return True
        family = socket.AF_UNSPEC
        pkt = nlpacket.Route(RTM_GETROUTE, self.__debug, self.__logger)
        pkt.flags = NLM_F_REQUEST | NLM_F_DUMP
        pkt.body = struct.pack('Bxxxii', family, 0, 0)
        pkt.build_message(self.__seqno.next(), self.__pid)
        return self.__tx_nlpacket(pkt)

    def decode_vrf_table(self, data, is_slave):
        """
        Until nlmanager can parse VRF stuff, lets do it here
        """

        (attr_len, attr_type) = struct.unpack('=HH', data[:4])
        attr_end = padded_length(attr_len)
        if not attr_end:
            return None

        data = data[4:attr_end]
        while data:
            (sub_attr_length, sub_attr_type) = struct.unpack('=HH', data[:4])
            sub_attr_end = padded_length(sub_attr_length)
            if not sub_attr_length:
                return None

            if ((not is_slave and sub_attr_type == Netlink.IFLA_INFO_DATA) or
                    (is_slave and
                     sub_attr_type == Netlink.IFLA_INFO_SLAVE_DATA)):
                sub_attr_data = data[4:sub_attr_end]

                while sub_attr_data:
                    (info_data_length, info_data_type) = struct.unpack(
                        '=HH', sub_attr_data[:4])
                    info_data_end = padded_length(info_data_length)
                    if not info_data_end:
                        return None
                    if ((not is_slave and
                         info_data_type == Netlink.IFLA_VRF_TABLE) or
                            (is_slave and
                             info_data_type == Netlink.IFLA_VRF_PORT_TABLE)):
                        return struct.unpack('=I', sub_attr_data[4:8])[0]

                    sub_attr_data = sub_attr_data[info_data_end:]

            data = data[sub_attr_end:]
        return None

    def decode_brinfo(self, data):
        """
        Until nlmanager can parse bridge's info_data stuff, lets do it here
        """

        value = {}
        (attr_len, attr_type) = struct.unpack('=HH', data[:4])
        attr_end = padded_length(attr_len)
        if not attr_end:
            return None

        data = data[4:attr_end]
        while data:
            (sub_attr_length, sub_attr_type) = struct.unpack('=HH', data[:4])
            sub_attr_end = padded_length(sub_attr_length)
            if not sub_attr_length:
                return value

            if sub_attr_type == Netlink.IFLA_INFO_DATA:
                sub_attr_data = data[4:sub_attr_end]

                while sub_attr_data:
                    (info_data_length, info_data_type) = struct.unpack(
                        '=HH', sub_attr_data[:4])
                    info_data_end = padded_length(info_data_length)
                    if not info_data_end:
                        return value
                    if info_data_type == Netlink.IFLA_BR_VLAN_FILTERING:
                        value[Netlink.IFLA_BR_VLAN_FILTERING] = (
                            struct.unpack('=B', sub_attr_data[4:5])[0]
                        )
                    elif info_data_type == \
                            Netlink.IFLA_BR_TOPOLOGY_CHANGE_DETECTED:
                        value[Netlink.IFLA_BR_TOPOLOGY_CHANGE_DETECTED] = (
                            struct.unpack('=B', sub_attr_data[4:5])[0]
                        )
                    elif info_data_type == Netlink.IFLA_BR_TOPOLOGY_CHANGE:
                        value[Netlink.IFLA_BR_TOPOLOGY_CHANGE] = (
                            struct.unpack('=B', sub_attr_data[4:5])[0])
                    sub_attr_data = sub_attr_data[info_data_end:]

            data = data[sub_attr_end:]

        return value

    def handle_netlink_msg(self, pkt, _):
        """ Parses an incoming RTNL message and places the result on a queue
        identified by the interface's index. Notifies the dispatcher by
        placing the interface's index on the "intfevent" queue.
        :param pkt: netlink message
        """
        offset = 0
        while offset < len(pkt):
            msg_len, msg_type, flags, seq, pid = (
                struct.unpack_from(self.__NLMSG_FMT,
                                   pkt,
                                   offset)
            )
            # 99% of the time when we see an ERROR the error code is zero
            # which means ACK
            possible_ack = False
            if msg_type == nlpacket.NLMSG_DONE:
                possible_ack = True
            elif msg_type == nlpacket.NLMSG_ERROR:
                error_code = int(
                    struct.unpack_from('>H', pkt, offset + self.__NLMSG_LEN)[0]
                )
                if error_code:
                    self.__logger.debug("RXed NLMSG_ERROR code %d", error_code)
                else:
                    possible_ack = True
            if possible_ack:
                self.__unacknowledged -= {(seq, pid)}
            if msg_type in self.__process_cbs:
                if msg_type in [RTM_NEWADDR, RTM_DELADDR]:
                    msg = nlpacket.Address(msg_type, self.__debug,
                                           self.__logger)
                elif msg_type in [RTM_NEWLINK, RTM_DELLINK]:
                    msg = nlpacket.Link(msg_type, self.__debug,
                                        self.__logger)
                elif msg_type in [RTM_NEWNEIGH, RTM_DELNEIGH]:
                    msg = nlpacket.Neighbor(msg_type, self.__debug,
                                            self.__logger)
                elif msg_type in [RTM_NEWROUTE, RTM_DELROUTE]:
                    msg = nlpacket.Route(msg_type, self.__debug,
                                         self.__logger)
                try:
                    msg.decode_packet(msg_len, flags, seq, pid,
                                      pkt[offset:offset + msg_len])
                except Exception:  # pylint: disable=broad-except
                    if self.__debug:
                        msg.dump()
                    self.__logger.error('Error decoding netlink msg.')
                else:
                    info = None
                    if msg_type in self.stats.ALL:
                        self.stats.increment(msg_type, 1)
                    if msg_type in [RTM_NEWADDR, RTM_DELADDR]:
                        info = Address(
                            msg.ifindex,
                            str(msg.get_attribute_value(Netlink.IFA_ADDRESS)),
                            str(msg.prefixlen),
                            msg.family
                        )
                    elif msg_type in [RTM_NEWLINK, RTM_DELLINK]:
                        ifla_af_spec = None
                        ifindex = msg.ifindex
                        ifname = msg.get_attribute_value(Netlink.IFLA_IFNAME)
                        # ifalias is a string, not a generic attr. Change it so
                        # that we can get the value without writing a decoder.
                        msg.attribute_to_class[Netlink.IFLA_IFALIAS] = (
                            'IFLA_IFALIAS', nlpacket.AttributeString)
                        ifalias = msg.get_attribute_value(Netlink.IFLA_IFALIAS)
                        mac_address = (
                            msg.get_attribute_value(Netlink.IFLA_ADDRESS)
                        )
                        admin_state = (
                            InterfaceState.UP if msg.is_up() else
                            InterfaceState.DOWN
                        )

                        flags = msg.get_flags_string()
                        if flags and 'IFF_RUNNING' in flags:
                            oper_state = InterfaceState.UP
                            down_reason = InterfaceDownReason.NONE
                        else:
                            oper_state = InterfaceState.DOWN
                            if 'IFF_PROTO_DOWN' in flags:
                                down_reason = InterfaceDownReason.PROTODOWN
                            elif 'IFF_LOWER_UP' not in flags:
                                down_reason = InterfaceDownReason.NO_CARRIER
                            elif 'IFF_DORMANT' in flags:
                                down_reason = InterfaceDownReason.DORMANT
                            else:
                                down_reason = ''

                        mtu = msg.get_attribute_value(Netlink.IFLA_MTU)
                        linkinfo = (
                            msg.attributes.get(Netlink.IFLA_LINKINFO, None)
                        )
                        if msg.family == socket.AF_BRIDGE:
                            kind = 'ignore'  # AF_BRIDGE doesn't provide kind
                            ifla_af_spec = msg.get_attribute_value(
                                Netlink.IFLA_AF_SPEC)
                        elif linkinfo is not None:
                            kind = linkinfo.value.get(Netlink.IFLA_INFO_KIND,
                                                      None)
                        else:
                            kind = None

                        master_ifindex = (
                            msg.get_attribute_value(Netlink.IFLA_MASTER)
                        )

                        stp_state = StpPortState.UNINIT
                        learning_en = False
                        arp_suppress_en = False
                        if msg.family == socket.AF_BRIDGE:
                            protinfo = msg.attributes.get(Netlink.IFLA_PROTINFO,
                                                          None)
                            if protinfo:
                                stp_state = protinfo.value.get(
                                    Netlink.IFLA_BRPORT_STATE,
                                    StpPortState.UNINIT)
                                learning_en = protinfo.value.get(
                                    Netlink.IFLA_BRPORT_LEARNING)
                                arp_suppress_en = protinfo.value.get(
                                    Netlink.IFLA_BRPORT_ARP_SUPPRESS)

                        iface_vlans = ''
                        access_vlan = 0
                        if ifla_af_spec:
                            for (x_type, x_value) in ifla_af_spec.iteritems():
                                begin_vlan_range = None
                                if x_type == Netlink.IFLA_BRIDGE_VLAN_INFO:
                                    for (vlan_flag, vlan_id) in x_value:
                                        if vlan_flag & Netlink.BRIDGE_VLAN_INFO_PVID:
                                            access_vlan = vlan_id
                                        elif vlan_flag & Netlink.BRIDGE_VLAN_INFO_RANGE_BEGIN:
                                            begin_vlan_range = vlan_id
                                        elif vlan_flag & Netlink.BRIDGE_VLAN_INFO_RANGE_END:
                                            if not begin_vlan_range:
                                                iface_vlans += ' %s' % vlan_id
                                            else:
                                                iface_vlans += ' %s-%s' % (
                                                    begin_vlan_range,
                                                    vlan_id)
                                            begin_vlan_range = None
                                        else:
                                            iface_vlans += ' %s' % vlan_id
                            iface_vlans = iface_vlans.strip()

                        table_id = RouteTable.DEFAULT
                        if master_ifindex and linkinfo:
                            slave_kind = linkinfo.value.get(
                                Netlink.IFLA_INFO_SLAVE_KIND
                            )
                            if slave_kind == LinkType.VRF:
                                table_id = self.decode_vrf_table(linkinfo.data,
                                                                 True)

                        if kind == LinkType.VLAN:
                            if Netlink.IFLA_INFO_DATA in linkinfo.value:
                                linkinfo_attrs = (
                                    linkinfo.value[Netlink.IFLA_INFO_DATA]
                                )
                                vlan_id = int(
                                    linkinfo_attrs[Netlink.IFLA_VLAN_ID]
                                )
                            else:
                                vlan_id = 0

                            parent_if = msg.get_attribute_value(
                                Netlink.IFLA_LINK)
                            info = Link(family=msg.family, ifindex=ifindex,
                                        master=master_ifindex, ifalias=ifalias,
                                        ifname=ifname, admin_state=admin_state,
                                        oper_state=oper_state,
                                        mac=mac_address, kind=kind, mtu=mtu,
                                        rt_table_id=table_id,
                                        vni=vlan_id, parent_if=parent_if,
                                        vlans=iface_vlans,
                                        down_reason=down_reason,
                                        access_vlan=access_vlan,
                                        stp_state=stp_state)
                        elif (kind == LinkType.VXLAN and
                                Netlink.IFLA_INFO_DATA in linkinfo.value):
                            linkinfo_attrs = (
                                linkinfo.value[Netlink.IFLA_INFO_DATA]
                            )
                            vni = int(linkinfo_attrs[Netlink.IFLA_VXLAN_ID])
                            localip = str(
                                linkinfo_attrs.get(Netlink.IFLA_VXLAN_LOCAL,
                                                   None)
                            )
                            dstport = (
                                linkinfo_attrs.get(Netlink.IFLA_VXLAN_PORT,
                                                   None)
                            )
                            info = Link(  # pylint: disable=redefined-variable-type
                                family=msg.family, ifindex=ifindex,
                                master=master_ifindex, ifname=ifname,
                                admin_state=admin_state, oper_state=oper_state,
                                mac=mac_address, kind=kind, mtu=mtu,
                                dstport=dstport, localip=localip, vni=vni,
                                rt_table_id=table_id, ifalias=ifalias,
                                vlans=iface_vlans, access_vlan=access_vlan,
                                learning_en=learning_en,
                                arp_suppress_en=arp_suppress_en,
                                stp_state=stp_state, down_reason=down_reason)

                        elif (kind == LinkType.VRF and
                              Netlink.IFLA_INFO_DATA in linkinfo.value):
                            table_id = self.decode_vrf_table(linkinfo.data,
                                                             False)
                            if table_id:
                                table_id = int(table_id)
                            else:
                                table_id = 0
                            info = Link(family=msg.family, ifindex=ifindex,
                                        master=master_ifindex, ifname=ifname,
                                        admin_state=admin_state,
                                        oper_state=oper_state,
                                        mac=mac_address, kind=kind, mtu=mtu,
                                        rt_table_id=table_id, ifalias=ifalias,
                                        down_reason=down_reason)
                        elif (kind == LinkType.BRIDGE and
                              Netlink.IFLA_INFO_DATA in linkinfo.value):
                            brinfo = self.decode_brinfo(linkinfo.data)
                            if Netlink.IFLA_BR_VLAN_FILTERING in brinfo:
                                is_vlan_filtering = brinfo[
                                    Netlink.IFLA_BR_VLAN_FILTERING]
                            else:
                                is_vlan_filtering = 0
                            info = Link(family=msg.family, ifindex=ifindex,
                                        master=master_ifindex, ifname=ifname,
                                        admin_state=admin_state,
                                        oper_state=oper_state,
                                        mac=mac_address, kind=kind, mtu=mtu,
                                        rt_table_id=table_id,
                                        is_vlan_filtering=is_vlan_filtering,
                                        vlans=iface_vlans, ifalias=ifalias,
                                        down_reason=down_reason,
                                        learning_en=learning_en,
                                        arp_suppress_en=arp_suppress_en,
                                        access_vlan=access_vlan,
                                        stp_state=stp_state)
                        else:
                            info = Link(family=msg.family, ifindex=ifindex,
                                        master=master_ifindex,
                                        ifname=ifname, admin_state=admin_state,
                                        oper_state=oper_state,
                                        mac=mac_address, kind=kind, mtu=mtu,
                                        rt_table_id=table_id, vlans=iface_vlans,
                                        access_vlan=access_vlan,
                                        stp_state=stp_state, ifalias=ifalias,
                                        learning_en=learning_en,
                                        arp_suppress_en=arp_suppress_en,
                                        down_reason=down_reason)
                    elif msg_type in [RTM_NEWNEIGH, RTM_DELNEIGH]:
                        mac=msg.get_attribute_value(Netlink.NDA_LLADDR)
                        if isinstance(mac, str) and mac.startswith('01:00:5e'):
                            # Skip multicast entries
                            offset += msg_len
                            continue

                        if ((msg.state & Netlink.NUD_REACHABLE) or
                                (msg.state & Netlink.NUD_STALE) or
                                (msg.state & Netlink.NUD_PERMANENT)):
                            state = NeighborState.UPDATE
                        elif ((msg.state & Netlink.NUD_DELAY) or
                              (msg.state & Netlink.NUD_FAILED) or
                              (msg.state & Netlink.NUD_INCOMPLETE)):
                            state = NeighborState.DELETE
                        else:
                            state = NeighborState.IGNORE

                        origin = False
                        if msg.state & Netlink.NUD_PERMANENT:
                            origin = True

                        is_static = msg.state & Netlink.NUD_NOARP
                        info = Neighbor(
                            ifindex=msg.ifindex,
                            ip=msg.get_attribute_value(Netlink.NDA_DST),
                            mac=msg.get_attribute_value(Netlink.NDA_LLADDR),
                            vlan=msg.get_attribute_value(Netlink.NDA_VLAN),
                            port=msg.get_attribute_value(Netlink.NDA_PORT),
                            state=state, origin=origin,
                            family=msg.family, flags=msg.flags,
                            is_static=is_static
                        )
                    elif msg_type in [RTM_NEWROUTE, RTM_DELROUTE]:
                        prefix = msg.get_prefix_string()
                        if prefix.startswith('ff02::'):
                            # Skip solicited node multicast addr 
                            offset += msg_len
                            continue
                        info = Route(
                            msg.route_type == RouteType.RTN_LOCAL,
                            prefix,
                            Route.get_nexthops(msg),
                            str(msg.protocol),
                            msg.route_type,
                            str(msg.get_attribute_value(
                                Netlink.RTA_PREFSRC)),
                            msg.family,
                            msg.get_attribute_value(Netlink.RTA_TABLE)
                        )
                    if info is not None:
                        self.__intfqueue[info.ifindex].put((msg_type, info))
                        self.__intfevent.put(info.ifindex)
            offset += msg_len

    def reset(self):
        self.__seqno = iter(xrange(10000))
        self.__unacknowledged = set()
        self.__intfqueue = collections.defaultdict(eventlet.Queue)
        while self.__running:
            eventlet.sleep(1)
        self.__intfevent = eventlet.Queue()
        if self.dispatcher_gt is not None:
            eventlet.kill(self.dispatcher_gt)
        self.stats.reset()

    def run(self):
        """ Spawns and returns the dispatcher thread.
        """
        self.dispatcher_gt = self.__pool.spawn(self.__dispatcher)
        return self.dispatcher_gt


class NetlinkCopy(object):
    """Provides utility functions to copy values from the netlink msg to
      the python redis object.
    """
    @staticmethod
    def copy_address(dst, src, idx_to_ifname):
        # Why is src.ifindex missing for SVI Interfaces?
        dst.ifname = idx_to_ifname.get(src.ifindex, 'None')
        dst.prefix = src.prefix
        dst.mask = src.mask
        dst.is_ipv6 = src.is_ipv6

    @staticmethod
    def copy_link(dst, src):
        dst.ifindex = src.ifindex
        dst.ifname = src.ifname
        dst.admin_state = src.admin_state
        dst.oper_state = src.oper_state
        if src.mac_addr is not None and is_valid_mac(str(src.mac_addr)):
            dst.mac_address = src.mac_addr
        dst.kind = src.kind or ''
        dst.mtu = src.mtu
        # We don't copy master
        dst.localip = src.localip
        dst.dstport = src.dstport
        dst.vni = src.vni
        dst.rt_table_id = src.rt_table_id
        dst.is_vlan_filtering = src.is_vlan_filtering
        if src.family == socket.AF_BRIDGE:
            dst.access_vlan = src.access_vlan
            dst.vlans = src.vlans
            dst.learning_en = bool(src.learning_en)
            dst.arp_suppress_en = src.arp_suppress_en
        # Don't copy parent_if or stp_state here

    @staticmethod
    def copy_link_always(dst, src):
        if not dst or not src:
            return

        dst.ifindex = src.ifindex
        dst.oper_state = src.oper_state
        if src.mac_addr is not None and is_valid_mac(str(src.mac_addr)):
            dst.mac_address = src.mac_addr
        if src.family == socket.AF_BRIDGE:
            dst.access_vlan = src.access_vlan
            dst.vlans = src.vlans
        dst.admin_state = src.admin_state
        dst.oper_state = src.oper_state
        dst.down_reason = src.down_reason
        if src.family == socket.AF_BRIDGE:
            dst.learning_en = bool(src.learning_en)
            if not src.arp_suppress_en:
                dst.arp_suppress_en = False
            else:
                dst.arp_suppress_en = True

    @staticmethod
    def copy_macfdb(dst, src, idx_to_ifname):
        # Why is src.ifindex missing for Bond Interfaces?
        dst.nexthop = idx_to_ifname.get(src.ifindex, 'None')
        if src.ip_address != 'None':
            dst.dst = src.ip_address
        else:
            dst.dst = ''
        if src.port is not None:
            dst.port = src.port
        else:
            dst.port = '0'
        if src.vlan is not None:
            dst.vlan = src.vlan
        else:
            dst.vlan = '0'
        dst.origin = src.origin
        if src.mac_address is not None and is_valid_mac(str(src.mac_address)):
            dst.mac_address = src.mac_address

        if src.flags & nlpacket.Neighbor.NTF_EXT_LEARNED:
            dst.is_remote = True
        else:
            dst.is_remote = False

        src.is_static = dst.is_static

    @staticmethod
    def copy_neighbor(dst, src, idx_to_ifname):
        # Why is src.ifindex missing for Bond Interfaces?
        dst.ifname = idx_to_ifname.get(src.ifindex, 'None')
        dst.ip_address = src.ip_address
        dst.is_ipv6 = src.is_ipv6
        if src.mac_address is not None and is_valid_mac(str(src.mac_address)):
            dst.mac_address = src.mac_address
        if src.flags & nlpacket.Neighbor.NTF_EXT_LEARNED:
            dst.is_remote = True
        else:
            dst.is_remote = False

    @staticmethod
    def copy_route(dst, src, idx_to_ifname, protos_map):
        dst.prefix = src.ip_address
        dst.nexthops = [
            (str(nexthop), idx_to_ifname.get(ifindex, None))
            for nexthop, ifindex in src.nexthops
        ]
        dst.protocol = protos_map.get(src.protocol, 'unknown')
        # IPv6 kernel hack fix
        if (src.is_ipv6 and src.route_type == RouteType.RTN_LOCAL and
                src.protocol == 0):
            dst.protocol = Protocol.KERNEL
        dst.origin = (src.origin or dst.protocol == Protocol.KERNEL or
                      dst.protocol == Protocol.STATIC)
        dst.src = src.src
        dst.is_ipv6 = src.is_ipv6
        if (src.is_ipv6 and src.route_type == RouteType.RTN_UNREACHABLE and
                len(dst.nexthops) == 1 and dst.nexthops[0][1] == 'lo'):
            dst.route_type = RouteType.RTN_UNICAST
        else:
            dst.route_type = src.route_type
        dst.rt_table_id = src.table
