# Copyright 2016, 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""ROM models
"""
import inspect
import time
import hashlib

from datetime import datetime
from netq_lib.orm import objects, HOSTNAME
from netq_lib.orm.memdb.models import MemModel
from netq_lib.orm.redisdb import Connection
from netq_lib.orm.redisdb.query import RedisQuery

# If you add any classes to this file that is not an object pushed to the
# backend DB, such as  _RedisModel and Connection, please add those classes
# to the ignored list in agent.py's reinitialize_db() function.

_KEY_TTL = 90                   # TTL liveness is 90s


class _RedisModel(objects.NetQModel):
    __memmodel = None
    query_cls = RedisQuery
    active = objects.Boolean(json_key='active')
    hostname = objects.String(json_key='hostname')
    timestamp = objects.Float(json_key='timestamp')
    deleted = objects.Boolean(json_key='deleted')

    def __init__(self):
        self.active = True
        self.deleted = False
        self.hostname = HOSTNAME
        self.timestamp = time.time()
        super(_RedisModel, self).__init__()

    @classmethod
    def exclude_fmt(cls):
        """Return the list of fields to ignore when doing a comparison to
        find the differences between two instances of an object.
        See BgpSession for an example usage"""
        return cls.timestamp, cls.active

    @classmethod
    def memmodel(cls):
        if cls.__memmodel is None:
            cls.__memmodel = type(cls.__name__, (MemModel, cls), {})
        return cls.__memmodel

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    @staticmethod
    def _get_connection_object(commit):
        if commit:
            return Connection.REDIS
        else:
            return Connection.PIPE

    def copy(self, src):
        for _, field in inspect.getmembers(self.__class__):
            if (isinstance(field, objects.Field) and
                    field.name != self.__class__.timestamp.name):
                src_attribute = getattr(src, field.name, None)
                if src_attribute is not None:
                    setattr(self, field.name, src_attribute)

    def delete(self, purge=False):
        '''Mark the object as deleted OR purge the entry from the backend'''
        if purge:
            cls = self.__class__
            connection = self._get_connection_object(True)
            connection.zremrangebyrank(cls.obj_to_key(self.__dict__),
                                       min=0, max=-1)
            connection.hdel(cls.SEPERATOR.join((cls._namespace, cls.__name__)),
                            cls.obj_to_key(self.__dict__))
            # Keeping the local cached around in the case of the agent is
            # probably fine. See comment in flush_entry() below too.
        else:
            obj = self.__class__()
            obj.copy(self)
            obj.active = False
            obj.deleted = True
            obj.save()

    def flush_entry(self):
        '''purge a single entry from the sorted set with specified score'''
        cls = self.__class__
        score = cls.obj_to_score(self.__dict__)
        key = cls.obj_to_key(self.__dict__)
        return Connection.REDIS.zremrangebyscore(key, min=score, max=score)
        # From an agent perspective, the local cached entry is still around.
        # Is this the right thing or the wrong thing ? If I just deleted an
        # entry, do I want to add back the same entry again ? I'd think not.
        # By keeping the cached entry around, nothing gets pushed unless
        # something changes, which seems logical.

    def get_objid(self):
        '''A unique for this object type/hostname ID'''
        return hashlib.md5(self.obj_to_key(self.__dict__)).hexdigest()

    def save_entry(self, connection, publish=False):

        cls = self.__class__
        mcls = cls.memmodel()
        key = cls.obj_to_key(self.__dict__)
        val = cls.obj_to_val(self.__dict__)
        score = cls.obj_to_score(self.__dict__)

        add = True
        delete = False
        old_obj = None

        if mcls is not None:
            old_objs = (
                list(mcls.query.range(start=0, end=10000, first=False,
                                      keys=[key]))
            )
            if old_objs:
                old_obj = old_objs[-1]

                if self.equals(old_obj):
                    if old_obj.active == self.active:
                        add = False
                    elif self.active and len(old_objs) >= 2:
                        delete = True

                for obj in old_objs[:-2]:
                    obj.delete_object()

        if delete:
            del_score = cls.obj_to_score(old_obj.__dict__)
            mcls.delete(key, del_score)
            connection.zremrangebyscore(key, del_score, del_score)
        elif add:
            mcls.save(key, score, val)
            connection.zadd(key, score, val)
            connection.hset(cls.SEPERATOR.join((
                cls._namespace, cls.__name__)),
                key, self.active)
            if publish:
                connection.publish(key, val)

    def counter_fmt(self):
        return None

    def counter_retention_policy(self):
        return None

    def counter_continuous_query(self):
        return None

    def diff(self, other):
        '''
        Returns dict { k:(self,other) } differences between self and other
        instance same class. Only finds deltas between common keys
        '''

        if other is None:
            raise ValueError

        if not type(self) is type(other):
            raise NotImplementedError

        # WARNING: Using PREFIX from objects.
        exclude_fields = {'i_' + field.name for field in self.exclude_fmt()}
        return {k: (v, other.__dict__.get(k))
                for k, v in self.__dict__.items()
                if k not in exclude_fields and other.__dict__.get(k) != v}


class Address(_RedisModel):
    ifname = objects.String(json_key='interface')
    prefix = objects.String(json_key='prefix')
    mask = objects.Integer(json_key='mask')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    vrf = objects.String(json_key='vrf')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname, cls.prefix,
                cls.mask, cls.is_ipv6, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.timestamp, cls.active, cls.deleted

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Address, self).save_entry(connection)


class BgpSession(_RedisModel):
    """Store BGP Session specific info
    """
    peer_name = objects.String(json_key='peerName')
    state = objects.String(json_key='peerState')
    peer_router_id = objects.String(json_key='bgpRouterId')
    peer_asn = objects.Long(json_key='peerAsn')
    peer_hostname = objects.String(json_key='peerHostname')
    asn = objects.Long(json_key='bgpAsn')
    reason = objects.String(json_key='reason')
    ipv4_pfx_rcvd = objects.Integer(json_key='ipv4PrefixRx')
    ipv6_pfx_rcvd = objects.Integer(json_key='ipv6PrefixRx')
    evpn_pfx_rcvd = objects.Integer(json_key='evpnPrefixRx')
    last_reset_time = objects.Float(json_key='lastResetTime')
    up_time = objects.Float(json_key='upTime')
    conn_estd = objects.Integer(json_key='connEstd')
    conn_dropped = objects.Integer(json_key='connDropped')
    upd8_rx = objects.Integer(json_key='upd8Rx')
    upd8_tx = objects.Integer(json_key='upd8Tx')
    vrfid = objects.Integer(json_key='vrfId')
    vrf = objects.String(json_key='vrf')
    tx_families = objects.Json(json_key='txFamilies')
    rx_families = objects.Json(json_key='rxFamilies')
    objid = objects.String(json_key='objId')

    @classmethod
    def exclude_fmt(cls):
        return (cls.timestamp, cls.active, cls.last_reset_time, cls.up_time,
                cls.upd8_tx, cls.upd8_rx)

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.peer_name, cls.asn, cls.vrf

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.peer_router_id, cls.peer_asn, cls.peer_hostname,
                cls.reason, cls.ipv4_pfx_rcvd, cls.ipv6_pfx_rcvd,
                cls.evpn_pfx_rcvd, cls.timestamp, cls.last_reset_time,
                cls.conn_estd, cls.conn_dropped, cls.upd8_rx, cls.vrfid,
                cls.upd8_tx, cls.up_time, cls.tx_families, cls.objid,
                cls.rx_families, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(BgpSession, self).save_entry(connection, publish=True)


class ClagSession(_RedisModel):
    role = objects.String(json_key='role')
    peer_role = objects.String(json_key='peerRole')
    peer_state = objects.Boolean(json_key='peerState')
    peer_if = objects.String(json_key='clagPeerIf')
    clag_sysmac = objects.String(json_key='sysMac')
    backup_ip_active = objects.Boolean(json_key='backupIpActive')
    backup_ip = objects.String(json_key='backupIp')
    single_bonds = objects.Json(json_key='singleBonds')
    dual_bonds = objects.Json(json_key='dualAttachedBonds')
    conflicted_bonds = objects.Json(json_key='conflictedBonds')
    proto_down_bonds = objects.Json(json_key='protoDownBonds')
    vxlan_anycast = objects.String(json_key='vxlanAnycast')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.clag_sysmac

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.peer_role, cls.role, cls.peer_state, cls.peer_if,
                cls.backup_ip, cls.backup_ip_active, cls.single_bonds,
                cls.dual_bonds, cls.timestamp, cls.conflicted_bonds,
                cls.vxlan_anycast, cls.proto_down_bonds, cls.active,
                cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ClagSession, self).save_entry(connection, publish=True)


class EvpnVniInfo(_RedisModel):
    '''Contains VNI info obtained from bgp evpn'''
    vni = objects.Integer(json_key='vni')
    origin_ip = objects.String(json_key='originIp')
    rd = objects.String(json_key='rd')
    export_rt = objects.Json(json_key='exportRT')
    import_rt = objects.Json(json_key='importRT')
    in_kernel = objects.Boolean(json_key='inKernel')
    adv_all_vni = objects.Boolean(json_key='advAllVni')
    adv_gw_ip = objects.String(json_key='advGatewayMacIp')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.vni

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.origin_ip, cls.rd, cls.export_rt, cls.import_rt,
                cls.in_kernel, cls.timestamp, cls.adv_all_vni, cls.adv_gw_ip,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EvpnVniInfo, self).save_entry(connection, publish=True)


class LnvSession(_RedisModel):
    role = objects.String(json_key='role')
    snd_ip = objects.String(json_key='sndIp')
    repl_mode = objects.String(json_key='replicationMode')
    rd_peers = objects.String(json_key='vxrdPeers')
    snd_peers = objects.Json(json_key='vxsndPeers')
    vnis = objects.Json(json_key='vnis')
    state = objects.String(json_key='state')
    version = objects.String(json_key='version')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.role

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.role, cls.snd_ip, cls.rd_peers, cls.snd_peers,
                cls.vnis, cls.state, cls.repl_mode, cls.version,
                cls.active, cls.deleted, cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(LnvSession, self).save_entry(connection, publish=True)


class Command(_RedisModel):
    key = objects.String(json_key='key')
    output = objects.String(json_key='output')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.key

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.output,

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        connection.zadd(Command.obj_to_key(self.__dict__),
                        Command.obj_to_score(self.__dict__),
                        Command.obj_to_val(self.__dict__))
        connection.hset(self.__class__.SEPERATOR.join((
            self.__class__._namespace,
            self.__class__.__name__)),
            Command.obj_to_key(self.__dict__),
            True)


class Link(_RedisModel):
    ifname = objects.String(json_key='interface')
    ifindex = objects.Integer(json_key='ifIndex')
    mac_address = objects.MacAddress(json_key='macAddress')
    admin_state = objects.String(json_key='adminState')
    oper_state = objects.String(json_key='operState')
    managed = objects.Boolean(json_key='managed')
    kind = objects.String(json_key='type')
    mtu = objects.Integer(json_key='mtu')
    master = objects.String(json_key='master')
    is_vlan_filtering = objects.Boolean(json_key='isVlanFiltering')
    vlans = objects.String(json_key='vlans')
    access_vlan = objects.Integer(json_key='pvid')
    dstport = objects.Integer(json_key='dstPort')
    localip = objects.String(json_key='localIp')
    vni = objects.Integer(json_key='vni')
    vrf = objects.String(json_key='vrf')
    rt_table_id = objects.Integer(json_key='routeTableId')
    parent_if = objects.String(json_key='parentIf')
    stp_state = objects.Integer(json_key='stpState')
    down_reason = objects.String(json_key='downReason')
    learning_en = objects.Boolean(json_key='learningEnabled')
    arp_suppress_en = objects.Boolean(json_key='arpSuppressionEnabled')
    objid = objects.String(json_key='objId')

    @classmethod
    def key_fmt(cls, key_type=None):
        """We need to query sometimes by VNI and so we stick VNI in the key.
        Its 0 for anything but Vxlan.
        """
        return (cls.__name__, cls.hostname, cls.ifname, cls.kind, cls.vni,
                cls.master)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.admin_state, cls.oper_state, cls.managed, cls.mtu,
                cls.ifindex, cls.is_vlan_filtering, cls.timestamp,
                cls.vlans, cls.access_vlan, cls.localip, cls.down_reason,
                cls.vrf, cls.rt_table_id, cls.parent_if, cls.stp_state,
                cls.mac_address, cls.dstport, cls.learning_en, cls.objid,
                cls.arp_suppress_en, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Link, self).save_entry(connection, publish=True)


class Description(_RedisModel):
    objtype = objects.String(json_key='objType')
    descrid = objects.String(json_key='descrId')
    description = objects.String(json_key='description')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.objtype, cls.descrid

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.description, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Description, self).save_entry(connection, publish=False)


class Port(_RedisModel):
    ifname = objects.String(json_key='interface')
    identifier = objects.String(json_key='module')
    speed = objects.String(json_key='speed')
    autoneg = objects.String(json_key='autoneg')
    state = objects.String(json_key='state')
    transreceiver = objects.String(json_key='transceiver')
    connector = objects.String(json_key='connector')
    vendor_name = objects.String(json_key='vendor')
    part_number = objects.String(json_key='partNo')
    serial_number = objects.String(json_key='serialNo')
    length = objects.String(json_key='length')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.identifier, cls.speed, cls.autoneg, cls.state,
                cls.transreceiver, cls.connector, cls.length, cls.vendor_name,
                cls.part_number, cls.serial_number, cls.active, cls.deleted,
                cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Port, self).save_entry(connection, publish=True)


class ProcNetDevInterface(_RedisModel):

    ifname = objects.String(json_key='interface')
    rx_bytes = objects.Integer(json_key='rxBytes')
    rx_drop = objects.Integer(json_key='rxDrop')
    rx_errs = objects.Integer(json_key='rxErrs')
    rx_frame = objects.Integer(json_key='rxFrame')
    rx_multicast = objects.Integer(json_key='rxMulticast')
    rx_packets = objects.Integer(json_key='rxPackets')
    tx_bytes = objects.Integer(json_key='txBytes')
    tx_carrier = objects.Integer(json_key='txCarrier')
    tx_colls = objects.Integer(json_key='txColls')
    tx_drop = objects.Integer(json_key='txDrop')
    tx_errs = objects.Integer(json_key='txErrs')
    tx_packets = objects.Integer(json_key='txPackets')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.rx_bytes, cls.rx_drop, cls.rx_errs,
                cls.rx_frame, cls.rx_multicast, cls.rx_packets,
                cls.tx_bytes, cls.tx_carrier, cls.tx_colls,
                cls.tx_drop, cls.tx_errs, cls.tx_packets,
                cls.active, cls.deleted, cls.timestamp)

    def counter_fmt(self):
        return {
            'measurement' : 'procnetdev',
            'time': datetime.fromtimestamp(self.timestamp).strftime('%Y-%m-%d %H:%M:%S.%f'),
            'tags' : {
                'hostname'  : self.hostname,
                'interface' : self.ifname,
            },
            'fields' : {
                'rxBytes'     : self.rx_bytes,
                'rxDrop'      : self.rx_drop,
                'rxErrs'      : self.rx_errs,
                'rxFrame'     : self.rx_frame,
                'rxMulticast' : self.rx_multicast,
                'rxPackets'   : self.rx_packets,
                'txBytes'     : self.tx_bytes,
                'txCarrier'   : self.tx_carrier,
                'txColls'     : self.tx_colls,
                'txDrop'      : self.tx_drop,
                'txErrs'      : self.tx_errs,
                'txPackets'   : self.tx_packets,
            },
        }

    def counter_retention_policy(self):
        # Delete the downsampled 5m data after one year
        return 'CREATE RETENTION POLICY "one_year" ON "netq" DURATION 52w REPLICATION 1'

    def counter_continuous_query(self):

        # Aggregate the raw 15s data from the previous 5m, this runs ever 5m
        return """
CREATE CONTINUOUS QUERY "cq_5m" ON "netq" BEGIN
  SELECT
    MEAN("rxBytes") AS "mean_rxBytes",
    MEAN("rxDrop") AS "mean_rxDrop",
    MEAN("rxErrs") AS "mean_rxErrs",
    MEAN("rxFrame") AS "mean_rxFrame",
    MEAN("rxMulticast") AS "mean_rxMulticast",
    MEAN("rxPackets") AS "mean_rxPackets",
    MEAN("txBytes") AS "mean_txBytes",
    MEAN("txCarrier") AS "mean_txCarrier",
    MEAN("txColls") AS "mean_txColls",
    MEAN("txDrop") AS "mean_txDrop",
    MEAN("txErrs") AS "mean_txErrs",
    MEAN("txPackets") AS "mean_txPackets"
  INTO "one_year"."downsampled_stats"
  FROM "procnetdev"
  GROUP BY time(5m)
END
"""

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ProcNetDevInterface, self).save_entry(connection, publish=False)


class Liveness(_RedisModel):

    def __str__(self):
        return self.hostname

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.hostname,

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Liveness.obj_to_key(self.__dict__)
        val = Liveness.obj_to_val(self.__dict__)
        score = Liveness.obj_to_score(self.__dict__)
        connection.zadd(key, score, val)
        connection.expire(key, _KEY_TTL)
        connection.hset(self.__class__.SEPERATOR.join((
            self.__class__._namespace,
            self.__class__.__name__)),
            key, True)


class Lldp(_RedisModel):
    ifname = objects.String(json_key='interface')
    peer_hostname = objects.String(json_key='peerHostname')
    peer_ifname = objects.String(json_key='peerInterface')
    lldp_peer_bridge = objects.Boolean(json_key='lldpPeerBridge')
    lldp_peer_router = objects.Boolean(json_key='lldpPeerRouter')
    lldp_peer_station = objects.Boolean(json_key='lldpPeerStation')
    lldp_peer_os = objects.String(json_key='lldpPeerOS')
    lldp_peer_osv = objects.String(json_key='lldpPeerOsVersion')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.peer_hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.peer_ifname, cls.lldp_peer_bridge, cls.lldp_peer_router,
                cls.lldp_peer_station, cls.lldp_peer_os, cls.lldp_peer_osv,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Lldp, self).save_entry(connection, publish=True)


class MacFdb(_RedisModel):
    mac_address = objects.MacAddress(json_key='macAddress')
    origin = objects.Boolean(json_key='origin')
    nexthop = objects.String(json_key='egressPort')
    vlan = objects.Integer(json_key='vlan')
    dst = objects.String(json_key='dst')
    port = objects.String(json_key='port')
    is_remote = objects.Boolean(json_key='isRemote')
    is_static = objects.Boolean(json_key='isStatic')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.mac_address, cls.vlan

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.origin, cls.nexthop, cls.dst, cls.port, cls.timestamp,
                cls.is_remote, cls.is_static, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(MacFdb, self).save_entry(connection)


class VxlanRemoteDest(_RedisModel):
    vni = objects.Integer(json_key='vni')
    rdst = objects.String(json_key='remoteDest')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.vni, cls.rdst

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.vni, cls.rdst, cls.active, cls.deleted, cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(VxlanRemoteDest, self).save_entry(connection, publish=True)


class MstpInfo(_RedisModel):
    """Store MSTP non-port-specific info for troubleshooting
    """
    bridge_name = objects.String(json_key='bridgeName')
    state = objects.Boolean(json_key='state')
    root_port_name = objects.String(json_key='rootPortName')
    root_bridge = objects.String(json_key='rootBridge')
    topo_chg_ports = objects.Json(json_key='topoChangePort')
    time_since_tcn = objects.Long(json_key='timeSinceLastTcn')
    topo_chg_cntr = objects.Long(json_key='toptChangeCntr')
    bridge_id = objects.String(json_key='bridgeId')
    edge_ports = objects.Json(json_key='edgePorts')
    network_ports = objects.Json(json_key='networkPorts')
    disputed_ports = objects.Json(json_key='disputedPorts')
    bpduguard_ports = objects.Json(json_key='bpduguardPorts')
    bpduguard_err_ports = objects.Json(json_key='bpduguardErrPorts')
    ba_inconsistent_ports = objects.Json(json_key='baInconsistentPorts')
    bpdufilter_ports = objects.Json(json_key='bpdufilterPorts')
    is_vlan_filtering = objects.Boolean(json_key='isVlanFiltering')
    ports = objects.Json(json_key='ports')

    @classmethod
    def exclude_fmt(cls):
        return cls.timestamp, cls.active, cls.time_since_tcn

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.bridge_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.root_port_name, cls.topo_chg_ports, cls.time_since_tcn,
                cls.topo_chg_cntr, cls.ports, cls.edge_ports, cls.state,
                cls.network_ports, cls.disputed_ports, cls.bpduguard_ports,
                cls.root_bridge, cls.bridge_id, cls.bpduguard_err_ports,
                cls.ba_inconsistent_ports, cls.bpdufilter_ports,
                cls.is_vlan_filtering, cls.active, cls.deleted,
                cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(MstpInfo, self).save_entry(connection, publish=True)


class OspfNbr(_RedisModel):
    peer_id = objects.String(json_key='peerId')
    peer_addr = objects.String(json_key='peerAddress')
    state = objects.String(json_key='state')
    ifname = objects.String(json_key='interface')
    is_ipv6 = objects.String(json_key='isIPv6')
    area = objects.String(json_key='area')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.peer_id, cls.area

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.timestamp, cls.ifname, cls.is_ipv6,
                cls.peer_addr, cls.area, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OspfNbr, self).save_entry(connection, publish=True)


class OspfIf(_RedisModel):
    ifname = objects.String(json_key='interface')
    is_unnumbered = objects.Boolean(json_key='isUnnumbered')
    if_up = objects.Boolean(json_key='ifUp')
    area = objects.String(json_key='area')
    network_type = objects.String(json_key='networkType')
    nbr_count = objects.Integer(json_key='nbrCount')
    nbr_adj_count = objects.Integer(json_key='nbrAdjCount')
    cost = objects.Integer(json_key='cost')
    is_passive = objects.Boolean(json_key='isPassive')
    mtu = objects.Integer(json_key='mtu')
    dead_time = objects.Integer(json_key='deadTime')
    rexmit_time = objects.Integer(json_key='rexmitTime')
    hello_time = objects.Integer(json_key='helloTime')
    router_id = objects.String(json_key='routerId')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.area

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.network_type, cls.timestamp, cls.nbr_count, cls.if_up,
                cls.nbr_adj_count, cls.is_unnumbered, cls.is_passive, cls.cost,
                cls.mtu, cls.dead_time, cls.rexmit_time, cls.hello_time,
                cls.router_id, cls.area, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OspfIf, self).save_entry(connection, publish=True)


class Neighbor(_RedisModel):
    ifname = objects.String(json_key='interface')
    ifindex = objects.Integer(json_key='ifIndex')
    ip_address = objects.String(json_key='ipAddress')
    mac_address = objects.MacAddress(json_key='macAddress')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    vrf = objects.String(json_key='vrf')
    is_remote = objects.Boolean(json_key='isRemote')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname,
                cls.ip_address, cls.mac_address, cls.is_ipv6, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.ifindex, cls.timestamp, cls.is_remote,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Neighbor, self).save_entry(connection)


class Node(_RedisModel):
    lastboot = objects.Float(json_key='lastBoot')
    sys_uptime = objects.Float(json_key='sysUptime')
    last_reinit = objects.Float(json_key='lastReinit')
    ntp_state = objects.String(json_key='ntpState')
    version = objects.String(json_key='version')

    def __str__(self):
        return self.hostname

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.lastboot, cls.sys_uptime, cls.last_reinit, cls.ntp_state, cls.version

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        key = Node.obj_to_key(self.__dict__)
        score = Node.obj_to_score(self.__dict__)
        val = Node.obj_to_val(self.__dict__)
        connection.zadd(key, score, val)
        connection.hset(self.__class__.SEPERATOR.join((
            self.__class__._namespace,
            self.__class__.__name__)),
            key, True)


class Route(_RedisModel):
    prefix = objects.String(json_key='prefix')
    protocol = objects.String(json_key='protocol')
    route_type = objects.Integer(json_key='routeType')
    origin = objects.Boolean(json_key='origin')
    nexthops = objects.Json(json_key='nextHops')
    src = objects.String(json_key='src')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    rt_table_id = objects.Integer(json_key='routeTableId')
    vrf = objects.String(json_key='vrf')

    @classmethod
    def key_fmt(cls, key_type=None):
        # The exact sequence of key fields here is assumed in the lpm LUA
        # script. Do not change one without changing the other. We need both
        # VRF and table in key to handle redistribute neighbor table. Also names
        # are constant across nodes, but table ids may vary from box to box. So,
        # save the routes with the VRF name as well as table id.
        return (cls.__name__, cls.hostname, cls.prefix, cls.route_type,
                cls.rt_table_id, cls.is_ipv6, cls.origin, cls.protocol, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.nexthops, cls.src, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Route, self).save_entry(connection)


class Services(_RedisModel):
    name = objects.String(json_key='service')
    is_enabled = objects.Boolean(json_key='isEnabled')
    is_active = objects.Boolean(json_key='isActive')
    is_monitored = objects.Boolean(json_key='isMonitored')
    status = objects.Integer(json_key='status')
    vrf = objects.String(json_key='vrf')
    pid = objects.String(json_key='pid')
    start_time = objects.Float(json_key='startTime')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.is_enabled, cls.is_active, cls.status, cls.is_monitored,
                cls.start_time, cls.pid,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Services, self).save_entry(connection, publish=True)


class OS(_RedisModel):
    name = objects.String(json_key='name')
    version = objects.String(json_key='fullVersion')
    version_id = objects.String(json_key='version')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.name, cls.version,
                cls.version_id)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(OS, self).save_entry(connection, publish=True)


class CPU(_RedisModel):
    arch = objects.String(json_key='arch')
    nos = objects.String(json_key='cores')
    model = objects.String(json_key='model')
    max_freq = objects.String(json_key='freq')
    mem_total = objects.String(json_key='memory')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.arch, cls.nos, cls.model,
                cls.max_freq, cls.mem_total)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(CPU, self).save_entry(connection, publish=True)


class ASIC(_RedisModel):
    vendor = objects.String(json_key='vendor')
    model = objects.String(json_key='model')
    model_id = objects.String(json_key='modelId')
    core_bw = objects.String(json_key='coreBW')
    ports = objects.String(json_key='ports')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vendor, cls.model,
                cls.model_id, cls.core_bw, cls.ports)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(ASIC, self).save_entry(connection, publish=True)


class Disk(_RedisModel):
    name = objects.String(json_key='name')
    size = objects.String(json_key='size')
    d_type = objects.String(json_key='type')
    vendor = objects.String(json_key='vendor')
    transport = objects.String(json_key='transport')
    rev = objects.String(json_key='rev')
    model = objects.String(json_key='model')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.size, cls.d_type,
                cls.vendor, cls.transport, cls.rev, cls.model)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Disk, self).save_entry(connection, publish=True)


class Board(_RedisModel):
    vendor = objects.String(json_key='vendor')
    model = objects.String(json_key='model')
    base_mac = objects.String(json_key='baseMac')
    part_number = objects.String(json_key='partNo')
    mfg_date = objects.String(json_key='mfgDate')
    serial_number = objects.String(json_key='serialNo')
    label_revision = objects.String(json_key='rev')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vendor, cls.model,
                cls.base_mac, cls.part_number, cls.mfg_date, cls.serial_number,
                cls.label_revision)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Board, self).save_entry(connection, publish=True)


class Memory(_RedisModel):
    vendor = objects.String(json_key='vendor')
    size = objects.String(json_key='size')
    speed = objects.String(json_key='speed')
    name = objects.String(json_key='name')
    m_type = objects.String(json_key='type')
    serial_number = objects.String(json_key='serialNo')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.size, cls.speed,
                cls.m_type, cls.vendor, cls.serial_number)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Memory, self).save_entry(connection, publish=True)


class Temp(_RedisModel):
    s_name = objects.String(json_key='name')
    s_desc = objects.String(json_key='description')
    s_input = objects.Float(json_key='temp')
    s_crit = objects.Integer(json_key='critical')
    s_lcrit = objects.Integer(json_key='lcritical')
    s_max = objects.Integer(json_key='max')
    s_min = objects.Integer(json_key='min')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.s_state, cls.s_prev_state, cls.s_input,
                cls.s_msg, cls.s_crit, cls.s_max, cls.s_min, cls.s_lcrit)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Temp, self).save_entry(connection, publish=True)


class Fan(_RedisModel):
    s_name = objects.String(json_key='name')
    s_desc = objects.String(json_key='description')
    s_input = objects.Integer(json_key='speed')
    s_max = objects.Integer(json_key='max')
    s_min = objects.Integer(json_key='min')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.s_state, cls.s_prev_state, cls.s_input,
                cls.s_msg, cls.s_max, cls.s_min)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Fan, self).save_entry(connection, publish=True)


class PSU(_RedisModel):
    s_name = objects.String(json_key='name')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.s_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.timestamp, cls.s_state, cls.s_prev_state, cls.s_msg

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(PSU, self).save_entry(connection, publish=True)


class EsxHost(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    cluster = objects.String(json_key='cluster')
    esxhost = objects.String(json_key='esxhost')
    pnic = objects.String(json_key='pnic')
    dvs = objects.String(json_key='dvs')
    pg = objects.String(json_key='pg')
    encaptype = objects.Integer(json_key='encaptype')
    encap = objects.Integer(json_key='encap')
    vmname = objects.String(json_key='vmname')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    mtu = objects.Integer(json_key='mtu')
    pnic_admin_state = objects.String(json_key='pnic_admin_state')
    pnic_oper_state = objects.String(json_key='pnic_oper_state')
    peer_hostname = objects.String(json_key='peer_hostname')
    peer_ifname = objects.String(json_key='peer_ifname')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.vcenter, cls.datacenter, cls.cluster, cls.esxhost,
                cls.pnic, cls.dvs, cls.pg, cls.encaptype, cls.encap)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.ip, cls.mac, cls.mtu, cls.pnic_admin_state,
                cls.pnic_oper_state, cls.peer_hostname, cls.peer_ifname,
                cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxHost, self).save_entry(connection)


class EsxVmAttr(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    vmname = objects.String(json_key='vmname')
    ostype = objects.String(json_key='ostype')
    vnic = objects.String(json_key='vnic')
    pg = objects.String(json_key='portgroup')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    esxhost = objects.String(json_key='esxhost')
    mem = objects.String(json_key='memory')
    disk = objects.String(json_key='disk')
    cpus = objects.String(json_key='numcpus')
    vmpath = objects.String(json_key='vmpath')
    state = objects.String(json_key='powerstate')
    annotation = objects.String(json_key='annotation')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vcenter, cls.datacenter,
                cls.vmname, cls.ostype, cls.vnic, cls.pg, cls.ip, cls.mac)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.esxhost, cls.mem, cls.disk, cls.cpus, cls.vmpath,
                cls.state, cls.annotation, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxVmAttr, self).save_entry(connection)


class EsxVmBind(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    cluster = objects.String(json_key='cluster')
    esxhost = objects.String(json_key='esxhost')
    vmname = objects.String(json_key='vmname')
    ostype = objects.String(json_key='ostype')
    state = objects.String(json_key='state')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vcenter, cls.datacenter,
                cls.cluster, cls.esxhost, cls.vmname, cls.ostype, cls.state)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(EsxVmBind, self).save_entry(connection)


class DockerContainer(_RedisModel):
    container_id = objects.String(json_key='containerId')
    gw = objects.String(json_key='gateway')
    image = objects.String(json_key='image')
    ip = objects.String(json_key='containerIp')
    mac = objects.String(json_key='mac')
    name = objects.String(json_key='containerName')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    prefix_len = objects.Integer(json_key='prefixLen')
    port_list = objects.String(json_key='portList')
    service_id = objects.String(json_key='serviceId')
    service_name = objects.String(json_key='serviceName')
    status = objects.String(json_key='status')
    start_time = objects.String(json_key='startTime')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.name, cls.image,
                cls.network_name, cls.ip, cls.mac,
                cls.service_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.container_id, cls.status,
                cls.network_id, cls.gw, cls.prefix_len,
                cls.port_list, cls.service_id,
                cls.start_time, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerContainer, self).save_entry(connection)


class DockerPortMap(_RedisModel):
    container_port = objects.String(json_key='containerPort')
    container_id = objects.String(json_key='containerId')
    container_ip = objects.String(json_key='containerIp')
    host_port = objects.String(json_key='hostPort')
    host_ip = objects.String(json_key='hostIp')
    image = objects.String(json_key='image')
    mac = objects.String(json_key='mac')
    name = objects.String(json_key='containerName')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    node_id = objects.String(json_key='nodeId')
    proto = objects.String(json_key='proto')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.name, cls.container_ip, cls.proto, cls.container_port,
                cls.host_ip, cls.host_port, cls.network_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.container_id, cls.network_id,
                cls.image, cls.mac, cls.node_id, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerPortMap, self).save_entry(connection)


class DockerSwarmCluster(_RedisModel):
    cluster_version = objects.String(json_key='clusterVersion')
    cluster_name = objects.String(json_key='clusterName')
    cluster_id = objects.String(json_key='clusterId')
    docker_version = objects.String(json_key='version')
    nodes = objects.String(json_key='workerNodes')
    num_nodes = objects.Integer(json_key='numNodes')
    num_managers = objects.Integer(json_key='numManagers')
    managers = objects.String(json_key='managerNodes')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.cluster_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.docker_version, cls.cluster_version, cls.cluster_id,
                cls.num_nodes, cls.num_managers, cls.managers,
                cls.timestamp, cls.nodes, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerSwarmCluster, self).save_entry(connection)


class DockerSwarmNode(_RedisModel):
    cluster_id = objects.String(json_key='clusterId')
    cluster_name = objects.String(json_key='clusterName')
    docker_version = objects.String(json_key='dockerVersion')
    node_id = objects.String(json_key='nodeId')
    node_name = objects.String(json_key='nodeName')
    node_state = objects.String(json_key='nodeState')
    plugins = objects.String(json_key='plugins')
    role = objects.String(json_key='role')
    availability = objects.String(json_key='availability')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.cluster_name, cls.node_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.docker_version, cls.cluster_id, cls.node_id,
                cls.node_state, cls.role, cls.plugins,
                cls.availability, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerSwarmNode, self).save_entry(connection)


class DockerNetwork(_RedisModel):
    container_list = objects.String(json_key='containerList')
    driver = objects.String(json_key='driver')
    host_binding = objects.String(json_key='hostBinding')
    ipv6_enabled = objects.Boolean(json_key='ipv6Enabled')
    ip_masquerade = objects.Boolean(json_key='ipMasq')
    ipam_driver = objects.String(json_key='ipamDriver')
    encrypted = objects.Boolean(json_key='encrypted')
    default_bridge = objects.Boolean(json_key='defaultBridge')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    subnet = objects.String(json_key='subnet')
    gateway = objects.String(json_key='gateway')
    parent_interface = objects.String(json_key='parentInterface')
    vxlan_id = objects.Integer(json_key='vxlanId')
    mtu = objects.Integer(json_key='mtu')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.network_name, cls.driver)


    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.gateway, cls.parent_interface, cls.vxlan_id, cls.network_id,
                cls.mtu, cls.host_binding, cls.ipv6_enabled, cls.ip_masquerade,
                cls.encrypted, cls.default_bridge, cls.ipam_driver, cls.subnet,
                cls.container_list, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerNetwork, self).save_entry(connection)


class DockerService(_RedisModel):
    image = objects.String(json_key='image')
    mode = objects.String(json_key='mode')
    parallelism = objects.Integer(json_key='parallelism')
    port_list = objects.String(json_key='portMapping')
    network_list = objects.String(json_key='networkList')
    replicas = objects.Integer(json_key='replicas')
    service_name = objects.String(json_key='serviceName')
    service_id = objects.String(json_key='serviceId')
    version = objects.Integer(json_key='version')
    vip = objects.String(json_key='virtualIp')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.service_name, cls.mode)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.image, cls.replicas, cls.parallelism, cls.service_id,
                cls.port_list, cls.network_list, cls.vip, cls.version,
                cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerService, self).save_entry(connection)


class DockerHost(_RedisModel):
    containers = objects.Integer(json_key='installed')
    containers_running = objects.Integer(json_key='running')
    docker_version = objects.String(json_key='version')
    images = objects.Integer(json_key='images')
    ip_forwarding = objects.Boolean(json_key='ipForwarding')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.docker_version)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.images, cls.containers, cls.containers_running,
                cls.ip_forwarding, cls.timestamp, cls.active, cls.deleted)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(DockerHost, self).save_entry(connection)


class License(_RedisModel):
    name = objects.String(json_key='name')
    state = objects.String(json_key='state')
    license = objects.String(json_key='license')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.state, cls.license, cls.timestamp

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(License, self).save_entry(connection, publish=True)

class Ntp(_RedisModel):
    current_server = objects.String(json_key='currentServer')
    stratum = objects.Integer(json_key='stratum')
    ntp_sync = objects.String(json_key='ntpSync')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.current_server, cls.stratum,
                cls.ntp_sync, cls.timestamp)

    def save(self, commit=False):
        connection = self._get_connection_object(commit)
        super(Ntp, self).save_entry(connection, publish=True)
